----------------------------------------------------------------------------------------------------
-- (c)2011 Altera Corporation. All rights reserved.
--
-- Altera products are protected under numerous U.S. and foreign patents,
-- maskwork rights, copyrights and other intellectual property laws.
--
-- This reference design file, and your use thereof, is subject to and governed
-- by the terms and conditions of the applicable Altera Reference Design License
-- Agreement (either as signed by you or found at www.altera.com).  By using
-- this reference design file, you indicate your acceptance of such terms and
-- conditions between you and Altera Corporation.  In the event that you do not
-- agree with such terms and conditions, you may not use the reference design
-- file and please promptly destroy any copies you have made.
--
-- This reference design file is being provided on an "as-is�" basis and as an
-- accommodation and therefore all warranties, representations or guarantees of
-- any kind (whether express, implied or statutory) including, without
-- limitation, warranties of merchantability, non-infringement, or fitness for
-- a particular purpose, are specifically disclaimed.  By making this reference
-- design file available, Altera expressly does not recommend, suggest or
-- require that this reference design file be used in combination with any
-- other product not provided by Altera.
----------------------------------------------------------------------------------------------------
-- Author		 : Peter Schepers (pscheper)
----------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

entity Decoder_66b64b is
PORT(
	Clk  			: in std_logic;
	Enable			: in std_logic;
	Reset	 		: in std_logic;
	din				: in std_logic_vector(65 downto 0);
	Bitslip			: out std_logic;
	dout	 		: out std_logic_vector(63 downto 0);
	kout			: out std_logic_vector(7 downto 0);	
	valid 			: out std_logic;
	Error			: buffer std_logic;
	Aligned			: out std_logic	
	);
END Decoder_66b64b;	


-- upper 2 bits are for framing 
-- bit[65:64] : "01" => data
--				"10" => control
--				"00" => not valid (used to detect framing)

architecture rtl of Decoder_66b64b is

CONSTANT TRESSHOLD_ALIGMENT_DEASSERT 	: unsigned(3 downto 0) := X"3";
CONSTANT LATENCY 						: unsigned(4 downto 0) := '0' & X"F"; 

signal count_Frame_OK 			: unsigned(5 downto 0);		
signal count_Frame_NotOK		: unsigned(5 downto 0);
signal slipcounter				: unsigned(6 downto 0);
signal Align					: std_logic;
signal Bitslip_i				: std_logic;
signal Bitslip_Qmin2				: std_logic;
signal Bitslip_Qmin1				: std_logic;
signal AlignChanged				: std_logic_vector(6 downto 0);
signal foundfirstcontrol		: std_logic;

component descrambler is
GENERIC
(
	WIDTH  : integer range 0 to 512 :=64
);
PORT
(
	clk						: in std_logic; 						
	arst					: in std_logic;	
	ena						: in std_logic;   						
	din						: in std_logic_vector(WIDTH-1 downto 0);	
	dout					: out std_logic_vector(WIDTH-1 downto 0)
 );
 end component;


begin


-----------------------------------------------------------------------------------------------------------
--	Align on valid 64/66 frame
-----------------------------------------------------------------------------------------------------------	

	process(Clk,Reset)
	begin
		if (Reset = '1') then
				Align 				<= '0';
				count_Frame_OK 	<= (OTHERS => '0');
				count_Frame_NotOK 	<= (OTHERS => '0');				
				slipcounter 		<= (OTHERS => '0');
				Bitslip					<= '0';
				AlignChanged 		<= (OTHERS => '0');
				Error		 		<= '0';
		elsif rising_edge(Clk) then

			if Enable = '1' then
	
	
				if slipcounter < LATENCY then  		-- keep in reset till after LATENCY parallel Clk cycles in order to let bitslip effect ripple through
					count_Frame_NotOK 	<=(Others => '0');
					count_Frame_OK 	<= (Others => '0');					

				else
					if (din(65 downto 64)  = "11") or (din(65 downto 64)  = "00") then
						count_Frame_NotOK  <= count_Frame_NotOK  + 1;
						count_Frame_OK <= (Others => '0');
					else
						count_Frame_OK <= count_Frame_OK + 1;
					end if;
				end if;
				

				if (count_Frame_OK(5) = '1') and (count_Frame_NotOK = 0)  and (foundfirstcontrol = '1') then -- Receive xx words with correct framing
					Align <= '1';
					
					assert(Align = '1')
					report "==========================================================================================================> Lane Aligned ..." 		severity note;
					AlignChanged <= (OTHERS => '0');
				
				elsif count_Frame_NotOK > TRESSHOLD_ALIGMENT_DEASSERT then -- Receive TRESSHOLD_ALIGMENT_DEASSERT or more not correct frames detected during a 256 cycle interval to declare Align lost : this is an indication the alignment has been lost															
					 Align <= '0'; 
				end if;
	
				if slipcounter(6) = '1' then 	-- Every xx Clk cycles check if Alignment is OK
					if  Align = '0' then
						Bitslip <= '1';
						if AlignChanged = "1000110" then  -- If no alignment can be achieved then there is something wrong with the input data
							AlignChanged <= (OTHERS => '0');
							Error	 <= '1';
							assert( Error = '1')
							report "==========================================================================================================> Can't find alignment" 		severity note;
							
						else
							AlignChanged <= AlignChanged + 1;
						end if;						
					else
						Bitslip <= '0';
					end if;	
					slipcounter <= (OTHERS => '0');
				else
					slipcounter <= slipcounter + 1;
					Bitslip <= '0';
				end if;
							

			end if;
			
		end if;
	end process;
	
Aligned <= Align;

-----------------------------------------------------------------------------------------------------------
--	DeScramble the data (note that bits 64 and 65 are not scrambled)
-----------------------------------------------------------------------------------------------------------	

descrambler_inst : descrambler
GENERIC MAP(
	WIDTH	=> 64
)
PORT MAP
(
	clk			=> Clk,						
	arst		=> Reset, 
	ena			=> Enable, 						
	din			=> din(63 downto 0),	
	dout		=> dout(63 downto 0)
);


-----------------------------------------------------------------------------------------------------------
--	Generate dout and kout
-----------------------------------------------------------------------------------------------------------	
	
	
process(Clk,Reset)
begin
	if Reset = '1' then
		valid 		<= '0';
		kout		<= (OTHERS => '0');
		foundfirstcontrol <= '0';
	elsif rising_edge(Clk) then
		valid 	<= Enable	; -- ReClock since Descrambler is taken one Clk cycle
		if Enable = '1' then	
			if din(65 downto 64) = "10" then
				foundfirstcontrol <= '1';
				kout	<= X"FF";		-- DELAY FOR ONE MORE CLOCK?
			else
				kout	<= X"00";
			end if;
		end if;
	end if;
end process;
	
	

			
end;



