----------------------------------------------------------------------------------------------------
-- (c)2007 Altera Corporation. All rights reserved.
--
-- Altera products are protected under numerous U.S. and foreign patents,
-- maskwork rights, copyrights and other intellectual property laws.
--
-- This reference design file, and your use thereof, is subject to and governed
-- by the terms and conditions of the applicable Altera Reference Design License
-- Agreement (either as signed by you or found at www.altera.com).  By using
-- this reference design file, you indicate your acceptance of such terms and
-- conditions between you and Altera Corporation.  In the event that you do not
-- agree with such terms and conditions, you may not use the reference design
-- file and please promptly destroy any copies you have made.
--
-- This reference design file is being provided on an "as-is�" basis and as an
-- accommodation and therefore all warranties, representations or guarantees of
-- any kind (whether express, implied or statutory) including, without
-- limitation, warranties of merchantability, non-infringement, or fitness for
-- a particular purpose, are specifically disclaimed.  By making this reference
-- design file available, Altera expressly does not recommend, suggest or
-- require that this reference design file be used in combination with any
-- other product not provided by Altera.
----------------------------------------------------------------------------------------------------
-- File          : rx_path_deskew.vhd
-- Author		 : Peter Schepers
----------------------------------------------------------------------------------------------------
--
-- Rx_Path_Deskew : 			
--							- Deskews the lanes and will reset the Rx receiver until maximum delay of 1 is seen
----------------------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use work.package_delaytype.all; -- package that define "Delay_Array" type

 
entity Rx_Path_Deskew is
GENERIC
	(
		NUMBER_OF_LANES	: integer := 5;	  	-- NUMBER_OF_LANES
		LANEWIDTH		: integer := 64		-- LANEWIDTH for transceiver
	);
PORT(
	Clock						: in std_logic;							
	Reset						: in std_logic;
	data_in						: in std_logic_vector(((NUMBER_OF_LANES * LANEWIDTH)-1) downto 0);
	ctrl_in						: in std_logic_vector((NUMBER_OF_LANES-1) DOWNTO 0);  		
	data_out						: out std_logic_vector(((NUMBER_OF_LANES * LANEWIDTH)-1) downto 0);
	rx_fifo_rd_en				: out STD_LOGIC_VECTOR((NUMBER_OF_LANES-1) DOWNTO 0); 	
	rx_enh_fifo_pempty		: in  STD_LOGIC_VECTOR((NUMBER_OF_LANES-1) DOWNTO 0); 
	Valid_Data					: out std_logic;
	Error							: out std_logic;
	Force_Rx_Reset				: out std_logic;		-- Not used
	Lane_Aligned				: buffer std_logic;
	Linkup						: buffer std_logic;
	XOFF_Received				: buffer std_logic;
	Rx_Error						: buffer std_logic;
	Delay							: buffer Delay_Array
	);
END Rx_Path_Deskew;	

	
architecture rtl of Rx_Path_Deskew is

type  DelayStateType is (ResetState, WaitForFifoFill, WaitForControl,WaitForLaneAligned,CheckDelay,WaitForControl_Validation,WaitForLaneAligned_Validation,CheckDelay_Validation,Success,ErrorState);
type   Rx_State_type	 is (Idle,Normal,XOFF_Received_from_Remote,Remote_Side_Aligned,Remote_Side_No_traffic,Remote_Side_No_traffic_and_XOFF_Request,Rx_State_Error);


signal DelayState 			: DelayStateType;
signal RX_State				: Rx_State_type;

signal Zeroes 					: std_logic_vector ((NUMBER_OF_LANES - 1) downto 0);
signal Ones 					: std_logic_vector ((NUMBER_OF_LANES - 1) downto 0);

signal Lane_Aligned_assertion	: std_logic;
signal Found_Alignment_Lane 	: std_logic_vector((NUMBER_OF_LANES-1) DOWNTO 0); 

signal countstate				: std_logic_vector(7 downto 0);
signal Delay_Combined		: std_logic_vector(((NUMBER_OF_LANES*3)-1) DOWNTO 0); 
signal Zeroes_3times			: std_logic_vector(((NUMBER_OF_LANES*3)-1) DOWNTO 0) := (OTHERS => '0');
signal ctrl_out				: std_logic_vector((NUMBER_OF_LANES-1) DOWNTO 0); 


begin

Generate_ZEROES_and_ONES:
FOR i IN 0 to NUMBER_OF_LANES-1 GENERATE
Ones(I) <= '1';
Zeroes(I) <= '0';
END GENERATE;

Generate_ZEROES_3times:
FOR i IN 0 to (NUMBER_OF_LANES*3)-1 GENERATE
Zeroes_3times(I) <= '0';
END GENERATE;


		
-----------------------------------------------------------------------------------------------------------
--	Delay Statemachine
-----------------------------------------------------------------------------------------------------------	

process (Clock,Reset)
	begin
		if Reset = '1' then
			DelayState <= ResetState;
		elsif Clock'event and Clock = '1' then

		     CASE DelayState	 IS
		         
					WHEN ResetState    		=> DelayState <= WaitForFifoFill;
					
					WHEN WaitForFifoFill		=>  if rx_enh_fifo_pempty = ZEROES then
															DelayState <= WaitForControl;
														 end if;
					
					WHEN WaitForControl 		=>  if ctrl_in(0) = '1' and data_in(7 downto 0) = X"BC"  then -- Wait to start checking for the alignment until first control character is received on lane 0
															DelayState <= WaitForLaneAligned;
														 end if;
														 
														 
					WHEN WaitForLaneAligned	=>	 if (Found_Alignment_Lane = Ones) then
															  DelayState <= CheckDelay;
														 end if;														 


					WHEN CheckDelay			=> if Delay_Combined = Zeroes_3times then -- check if Delay is 0 on all lanes , otherwise go back to WaitForControl
																DelayState <= Success;
														else
																DelayState <= WaitForControl;
														end if;
														
--					WHEN Trigger_Rx_Reset 	=>	if countstate = X"FF" then -- This should never be reached as it will be reset during this time, if it does there is an error
--																DelayState <= ErrorState;
--														end if;
					WHEN Success				=> if ctrl_in(0) = '1' and data_in(7 downto 0) = X"BC"  then -- Wait to start checking for the alignment until first control character is received on lane 0
															DelayState <= WaitForControl_Validation ;
														 end if;

					WHEN WaitForControl_Validation 		=>  if ctrl_in(0) = '1' and data_in(7 downto 0) = X"BC"  then -- Wait to start checking for the alignment until first control character is received on lane 0
															DelayState <= WaitForLaneAligned_Validation;
														 end if;
														 

					WHEN WaitForLaneAligned_Validation	=>	 if (Found_Alignment_Lane = Ones) then
															  DelayState <= CheckDelay_Validation;
														 end if;														 


					WHEN CheckDelay_Validation			=> if Delay_Combined = Zeroes_3times then -- check if Delay is 1 on all lanes , otherwise there has been a misalignment and should go to the errorstate
																DelayState <= Success;
														else
																DelayState <= ErrorState;
														end if;	
																	
					WHEN ErrorState			=> --DelayState <= ResetState;
														DelayState <= ErrorState;	

					When OTHERS					=> DelayState <= ErrorState;
				end case;	
						
		end if;
end process;

process (Clock,Reset)
	begin
		if Reset = '1' then
			FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP
				Delay(I) <= (OTHERS => '1');
			END LOOP;
			ctrl_out							<= (OTHERS => '0');
			Lane_Aligned 					<= '0';
			--Valid_Data						<= '0';
			Lane_Aligned_assertion 		<= '0';
			Error								<= '0';
			countstate						<= (OTHERS => '0');
			Force_Rx_Reset					<= '0';
			rx_fifo_rd_en					<= (OTHERS => '1');
			FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP
				Found_Alignment_Lane(I) <= '0';
			END LOOP;
		elsif Clock'event and Clock = '1' then

		     CASE DelayState	 IS
		         
					WHEN ResetState    		=> 
														FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP
															Delay(I) <= (OTHERS => '1');
														END LOOP;
														ctrl_out							<= (OTHERS => '0');
														Lane_Aligned 					<= '0';
														--Valid_Data						<= '0';
														Lane_Aligned_assertion 		<= '0';
														Error								<= '0';
														countstate						<= (OTHERS => '0');
														Force_Rx_Reset					<= '0';
														rx_fifo_rd_en					<= (OTHERS => '1');
														FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP
															Found_Alignment_Lane(I) <= '0';
														END LOOP;					

					WHEN WaitForFifoFill    =>  rx_fifo_rd_en					<= (OTHERS => '0');														
					
					WHEN WaitForControl 		=>  FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP	
															rx_fifo_rd_en(I) <= '1';	
															 END LOOP;	
														
														-- Reset the delay
														FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP
															Delay(I) <= (OTHERS => '1');
														END LOOP;														
															 
					WHEN WaitForLaneAligned	=>	 FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP				
																if ctrl_in(I) = '1' and data_in((LANEWIDTH*(I+1)-1) downto (LANEWIDTH*I)) = X"7C7C7C7C7C7C7C7C"  then
																		Found_Alignment_Lane(I) <= '1';
																end if;
														 END LOOP;
														 FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP	
															rx_fifo_rd_en(I) <= '1';	
															 END LOOP;

															FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP				
																if Found_Alignment_Lane(I) = '1' then
																	Delay(I) <= Delay(I) + 1;
																end if;
															END LOOP;	

															if (Found_Alignment_Lane = Ones) then 
																	FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP
																		Found_Alignment_Lane(I) <= '0';
																	END LOOP;		
															end if;	
															
					WHEN CheckDelay			=> 
														FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP
															if Delay(I) > 0 then			-- If delay is bigger than zero then do not read out the FIFO for one clock cycle in order to delay it for one clock
																rx_fifo_rd_en(I) <= '0';
															else
																rx_fifo_rd_en(I) <= '1';
															end if;
														END LOOP;
														

														
														
--					WHEN Trigger_Rx_Reset 	=>	if countstate = X"FF" then 
--															countstate <= (OTHERS => '0');
--														else
--															countstate <= countstate + 1;
--														end if;
--														Force_Rx_Reset <= '1';

					WHEN Success				=> FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP	
															rx_fifo_rd_en(I) <= '1';	
														END LOOP;
															 
														Lane_Aligned <= '1';
														if (Lane_Aligned = '1') then 
															Lane_Aligned_assertion <= '1';
															assert(Lane_Aligned_assertion = '1')
															report "==========================================================================================================> All lanes Aligned ..." 		severity note;										
														end if;
														
					WHEN WaitForControl_Validation 		=>  FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP	
															rx_fifo_rd_en(I) <= '1';	
															 END LOOP;	
														
														-- Reset the delay
														FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP
															Delay(I) <= (OTHERS => '1');
														END LOOP;															
														
					WHEN WaitForLaneAligned_Validation	=>	 FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP				
																if ctrl_in(I) = '1' and data_in((LANEWIDTH*(I+1)-1) downto (LANEWIDTH*I)) = X"7C7C7C7C7C7C7C7C"  then
																		Found_Alignment_Lane(I) <= '1';
																end if;
														 END LOOP;
														 FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP	
															rx_fifo_rd_en(I) <= '1';	
															 END LOOP;

															FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP				
																if Found_Alignment_Lane(I) = '1' then
																	Delay(I) <= Delay(I) + 1;
																end if;
															END LOOP;	

															if (Found_Alignment_Lane = Ones) then 
																	FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP
																		Found_Alignment_Lane(I) <= '0';
																	END LOOP;		
															end if;	
															
--														if (ctrl_in = Zeroes) and (Lane_Aligned = '1') then
--																Valid_Data <= '1';
--														else
--																Valid_Data <= '0';
--														end if;		

					WHEN CheckDelay_Validation			=> NULL;
--														FOR i IN 0 to (NUMBER_OF_LANES-1) LOOP
--															if Delay(I) > 0 then			-- If delay is bigger than zero then do not read out the FIFO for one clock cycle in order to delay it for one clock
--																rx_fifo_rd_en(I) <= '0';
--															else
--																rx_fifo_rd_en(I) <= '1';
--															end if;
--														END LOOP;
														
					WHEN ErrorState			=> Error <= '1';
					
					When OTHERS					=> Error <= '1';
				end case;	
						
		end if;
end process;

data_out <= data_in;
--Valid_Data <= '1' when ((ctrl_in = Zeroes) and (Lane_Aligned = '1')) else '0';
Valid_Data <= '1' when ((ctrl_in /= Ones) and (Lane_Aligned = '1')) else '0';
-- Modified to allow for glitches on ctrl_in to not bring down the entire link for a long time
-- Discovered during testing with FEC.

Generate_Delay_Combined_MSB:
FOR i IN 0 to (NUMBER_OF_LANES-1) GENERATE
	Delay_Combined((3*I)+2 downto (3*I)) <=  Delay(I)(2 downto 0); 		
END GENERATE;		

--DataOut_Valid 	<= '1' when (Ctrl_Aligned(0) = '0' and Reset_Rx_Path = '0') else '0';	




-----------------------------------------------------------------------------------------------------------
--	Rx_State State Machine
-----------------------------------------------------------------------------------------------------------	

	process (Clock,Reset)
	begin
		if Reset = '1' then
				Rx_State <= Idle;	
		elsif Clock'event and Clock = '1' then

		
			  if (ctrl_in = Ones) and (Lane_Aligned = '1') then 
			  
				  CASE data_in(15 downto 0) IS
				  
					  WHEN X"7C7C" => Rx_State <= Normal;  -- K28.3/K28.3
												assert (Rx_State = Normal)
												report "==========================================================================================================> Rx State Normal ..." 		severity note;
					  
					  WHEN X"FD7C" => Rx_State <= Remote_Side_Aligned; -- K29.7/K28.3
												assert (Rx_State = Remote_Side_Aligned)
												report "==========================================================================================================> Remote_Side_Aligned ..." 		severity note;
					  
					  WHEN X"5C7C" => Rx_State <= XOFF_Received_from_Remote; -- K28.2/K28.3
												assert (Rx_State = XOFF_Received_from_Remote)
												report "==========================================================================================================> XOFF_Received_from_Remote ..." 		severity note;

					  WHEN X"3C7C" => Rx_State <= Remote_Side_No_traffic; -- K28.1/K28.3
												assert (Rx_State = Remote_Side_No_traffic)
												report "==========================================================================================================> Remote_Side_No_traffic ..." 		severity note;

					  WHEN X"9C7C" => Rx_State <= Remote_Side_No_traffic_and_XOFF_Request; -- K28.4/K28.3
												assert (Rx_State = Remote_Side_No_traffic_and_XOFF_Request)
												report "==========================================================================================================> Remote_Side_No_traffic_and_XOFF_Request ..." 		severity note;
												
					  WHEN X"1CBC" => Rx_State <= Rx_State;
					  WHEN OTHERS  => Rx_State <= Idle;
												assert (Rx_State = Idle)
												report "==========================================================================================================> Rx_State = Idle ..." 		severity note;
					  					  
				  end case;
			 end if;
			  							
		end if;
   	end process;
		
	process (Clock,Reset)
	begin
		if Reset = '1' then
					Linkup <= '0';
					XOFF_Received <= '0';
					Rx_Error <= '0';
		elsif Clock'event and Clock = '1' then

		     CASE Rx_State IS
		         WHEN Idle    				=>   Linkup <= '0';
														  XOFF_Received <= '0';
														  
					WHEN Normal					=>   Linkup <= '0';
														  XOFF_Received <= '0';
					WHEN XOFF_Received_from_Remote => Linkup <= '1';
														   XOFF_Received <= '1';	
--															assert (XOFF_Received = '1')
--															report "==========================================================================================================> XOFF Received remote side ..." 		severity note;
															
					WHEN Remote_Side_Aligned => 	Linkup <= '1';
															assert (Linkup = '1')
															report "==========================================================================================================> Linkup ..." 		severity note;

														   XOFF_Received <= '0';																

					WHEN Remote_Side_No_traffic => 	Linkup <= '1';

														   XOFF_Received <= '0';																

					WHEN Remote_Side_No_traffic_and_XOFF_Request => 	Linkup <= '1';

														   XOFF_Received <= '1';																
														  
					When OTHERS					=> 	Linkup <= '0';
														   XOFF_Received <= '0';
															Rx_Error <= '1';
															assert (Rx_Error = '1')
															report "==========================================================================================================> Rx State Error ..." 		severity note;

															
														
															
				end case;								
		end if;
   	end process;	
		

					
end;



